// MotorController.cpp
#include "MotorController.h"

const MotorConfig defaultConfig = {
    .minPWM = 0,
    .maxPWM = 255,
    .inverted = true,
    .responseSpeed = 50,
    .lookupTable = {0, 10, 20, 30, 40, 50, 60, 70, 80, 100} // Percent values for 0% to 100%
};

MotorController::MotorController(int motorPin)
    : motorPWMPin(motorPin), currentSpeed(0), targetSpeed(0)
{
    loadDefaults();
    pinMode(motorPWMPin, OUTPUT);
    stopMotor();
}

void MotorController::loadDefaults()
{
    config = defaultConfig;
}

void MotorController::setTargetSpeed(int targetSpeedPercent)
{
    if (targetSpeed != targetSpeedPercent)
    {
        Serial.print("MC: set targetSpeed to: ");
        Serial.println(targetSpeed);
    }
    targetSpeed = constrain(targetSpeedPercent, 0, 100);
}

void MotorController::maintain()
{
    targetSpeed = constrain(targetSpeed, 0, 100);          // Ensure targetSpeed is valid
    int adjustedSpeed = applyDynamicResponse(targetSpeed); // Dynamic adjustment in percentage

    // Interpolate between lookup table values
    int lowerIndex = map(adjustedSpeed, 0, 100, 0, MotorConfig::lookupTableSize - 1);
    int upperIndex = min(lowerIndex + 1, MotorConfig::lookupTableSize - 1);
    int lowerPercent = config.lookupTable[lowerIndex];
    int upperPercent = config.lookupTable[upperIndex];

    float interpolationFactor = (adjustedSpeed - lowerIndex * (100 / (MotorConfig::lookupTableSize - 1))) / (100.0 / (MotorConfig::lookupTableSize - 1));
    int interpolatedPWM = map(lowerPercent + interpolationFactor * (upperPercent - lowerPercent), 0, 100, config.minPWM, config.maxPWM);

    setPWM(interpolatedPWM);      // Apply final interpolated PWM value
    currentSpeed = adjustedSpeed; // Update currentSpeed in percentage
}

int MotorController::getSpeed() const
{
    return currentSpeed;
}

void MotorController::stopMotor()
{
    setTargetSpeed(0);
    maintain();
}

MotorConfig MotorController::getConfig() const
{
    return config;
}

void MotorController::setConfig(const MotorConfig &newConfig)
{
    config = newConfig;
}

void MotorController::setPWM(uint8_t pwmValue)
{
    int constrainedPWM = constrain(pwmValue, config.minPWM, config.maxPWM);
    analogWrite(motorPWMPin, config.inverted ? 255 - constrainedPWM : constrainedPWM);
}

int MotorController::applyDynamicResponse(int targetPercent)
{
    // Intern in Promille arbeiten für höhere Genauigkeit
    static int smoothedPromille = 0;
    static unsigned long lastUpdateTime = millis();

    unsigned long currentTime = millis();
    unsigned long elapsedTime = currentTime - lastUpdateTime;

    if (elapsedTime == 0)
        return smoothedPromille / 10; // Rückgabe in Prozent

    // Zielwert in Promille berechnen
    int targetPromille = targetPercent * 10;

    // Dynamische Reaktionsgeschwindigkeit in Promille pro Sekunde
    int maxChangePerMs = config.responseSpeed * 10; // Konfigurierbarer Faktor

    // Berechne maximale Änderung basierend auf vergangener Zeit
    int maxDelta = (maxChangePerMs * elapsedTime) / 1000;

    // Tatsächliche Änderung begrenzen
    int delta = targetPromille - smoothedPromille;
    if (delta > maxDelta)
        delta = maxDelta;
    else if (delta < -maxDelta)
        delta = -maxDelta;

    // Smoothed Promille aktualisieren
    smoothedPromille += delta;
    lastUpdateTime = currentTime;

    // Debug-Ausgabe
    Serial.print("TargetPromille: ");
    Serial.print(targetPromille);
    Serial.print(", Delta: ");
    Serial.print(delta);
    Serial.print(", SmoothedPromille: ");
    Serial.println(smoothedPromille);

    // Rückgabe in Prozent (gerundet)
    return (smoothedPromille + 5) / 10; // Rundung durch Addition von 5 vor Division
}
