// PedalController.cpp
#include "PedalController.h"

const int minMin = 50;   // Minimum allowed raw value for calibration
const int maxMin = 200;  // Maximum allowed raw value for calibration
const int minMax = 800;  // Minimum allowed maximum value for calibration
const int maxMax = 1000; // Maximum allowed raw value for calibration

const PedalConfig defaultConfig = {
    .minRaw = 0,
    .maxRaw = 1023,
    .calibrated = false
    };

PedalController::PedalController(int pin)
    : pedalPin(pin)
{
    loadDefaults();
    pinMode(pedalPin, INPUT);
}

void PedalController::loadDefaults()
{
    config = defaultConfig;
}

int PedalController::applySmoothing(int rawValue)
{
    filteredValue = (filteredValue * (config.filterStrength - 1) + rawValue) / config.filterStrength;
    return filteredValue;
}

CalibrationState PedalController::autoCalibrate(bool reset)
{
    static int step = 0;
    static unsigned long lastChangeTime = 0;
    static int previousValue = 0;
    static CalibrationState errorState = IDLE;

    if (reset)
    {
        step = 0;
        config.minRaw = 1023;
        config.maxRaw = 0;
        lastChangeTime = millis();
        config.calibrated = false;
        errorState = IDLE;
        Serial.println("Calibration reset.");
        return IDLE;
    }

    unsigned long currentTime = millis();
    int rawValue = applySmoothing(analogRead(pedalPin));

    switch (step)
    {
    case 0: // Initial state: wait for pedal release
        if (rawValue == previousValue && currentTime - lastChangeTime > 2000)
        {
            if (rawValue >= minMin && rawValue <= maxMin)
            {
                config.minRaw = rawValue;
                step++;
                lastChangeTime = currentTime;
                errorState = RUNNING;
            }
            else
            {
                errorState = ERROR;
            }
        }
        else if (rawValue != previousValue)
        {
            lastChangeTime = currentTime;
        }
        break;

    case 1: // Wait for pedal full press
        if (rawValue == previousValue && currentTime - lastChangeTime > 2000)
        {
            if (rawValue >= minMax && rawValue <= maxMax)
            {
                config.maxRaw = rawValue;
                step++;
                lastChangeTime = currentTime;
                errorState = RUNNING;
            }
            else
            {
                errorState = ERROR;
            }
        }
        else if (rawValue != previousValue)
        {
            lastChangeTime = currentTime;
        }
        break;

    case 2: // Validate calibration
        if (config.maxRaw > config.minRaw + 100)
        { // Ensure valid range
            step++;
            config.calibrated = true;
            errorState = SUCCESS;
        }
        else
        {
            errorState = ERROR;
        }
        break;

    default:
        errorState = IDLE;
        step = 0;
        break;
    }

    previousValue = rawValue;
    return errorState;
}

int PedalController::getPedal()
{
    if (!config.calibrated || autoCalibrate(false) != IDLE)
    {
        return 0;
    }

    int rawValue = applySmoothing(analogRead(pedalPin));
    if (rawValue < config.minRaw || rawValue > config.maxRaw)
    {
        return 0; // Out of bounds
    }

    return map(rawValue, config.minRaw, config.maxRaw, 0, 100);
}

CalibrationState PedalController::getStatus()
{
    return autoCalibrate(false);
}

int PedalController::getRawValue()
{
    return analogRead(pedalPin);
}

PedalConfig PedalController::getConfig()
{
    return config;
}

void PedalController::setConfig(const PedalConfig &newConfig)
{
    config = newConfig;
}
