#include <Arduino.h>
#include <EEPROM.h>
#include <Wire.h>
#include <Adafruit_SSD1306.h>
#include "MotorController.h"
#include "PedalController.h"

// OLED display setup
#define OLED_RESET -1
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// Pin definitions
const int hallSensor = 2;    // Hall sensor input
const int encoderPinA = 3;   // Encoder pin A
const int encoderPinB = 4;   // Encoder pin B
const int encoderButton = 5; // Encoder button
const int motorPin = 9;      // Motor PWM pin
const int pedalPin = A0;     // Pedal input pin

// Global objects
MotorController motor(motorPin); // Motor control object
PedalController pedal(pedalPin); // Pedal input object

// Combined settings structure
struct Settings
{
  MotorConfig motorConfig;
  PedalConfig pedalConfig;
  uint8_t checksum;
};

Settings settings;
bool autoCalibrating = false;
bool serialControlEnabled = false;
CalibrationState lastCalibrationState = IDLE;

// Function prototypes
void loadSettings();
void saveSettings();
uint8_t calculateChecksum(const Settings &s);
void handleEncoder();
void handleEncoderButton();
void updateDisplay();
void handleSerialInput();
void configurePWMFrequency();
void stopMotorAtNeedleUp();
void printHelp();

void setup()
{
  // Initialize pins
  pinMode(hallSensor, INPUT);
  pinMode(encoderPinA, INPUT_PULLUP);
  pinMode(encoderPinB, INPUT_PULLUP);
  pinMode(encoderButton, INPUT_PULLUP);

  pinMode(LED_BUILTIN, OUTPUT);

  // Configure PWM frequency
  configurePWMFrequency();

  // Attach interrupts for encoder
  attachInterrupt(digitalPinToInterrupt(encoderPinA), handleEncoder, CHANGE);
  attachInterrupt(digitalPinToInterrupt(encoderPinB), handleEncoder, CHANGE);
  attachInterrupt(digitalPinToInterrupt(encoderButton), handleEncoderButton, FALLING);

  // Initialize serial and load settings
  Serial.begin(9600);
  loadSettings();

  // Initialize OLED display
  //  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
  //      Serial.println(F("SSD1306 allocation failed"));
  //  }
  //  display.clearDisplay();
  //   display.display();

  printHelp();
}

void loop()
{
  // Handle serial input for debugging and manual control
  handleSerialInput();
  motor.maintain();

  if (autoCalibrating)
  {
    CalibrationState currentState = pedal.autoCalibrate(false);
    if (currentState != lastCalibrationState)
    {
      lastCalibrationState = currentState;
      switch (currentState)
      {
      case RUNNING:
        Serial.println("Calibration running...");
        break;
      case SUCCESS:
        Serial.println("Auto-calibration completed.");
        break;
      case ERROR:
        Serial.println("Auto-calibration failed.");
        break;
      default:
        break;
      }
    }
    motor.setTargetSpeed(0); // Ensure motor stays off during calibration
  }
  else if (serialControlEnabled)
  {
    // Serial control active, do nothing here; speed controlled via serial commands
  }
  else
  {
    int pedalValue = pedal.getPedal();
    motor.setTargetSpeed(pedalValue);
  }

  // Update display
  //   updateDisplay();

  delay(10); // Short delay for stability
}

void stopMotorAtNeedleUp()
{
  motor.stopMotor();
  delay(100); // Stabilize
}

void loadSettings()
{
  EEPROM.get(0, settings);

  // Validate checksum
  if (calculateChecksum(settings) != settings.checksum)
  {
    Serial.println("Invalid settings checksum, loading defaults.");
    motor.loadDefaults();
    pedal.autoCalibrate(true);

    settings.motorConfig = motor.getConfig();
    settings.pedalConfig = pedal.getConfig();
    settings.checksum = calculateChecksum(settings);

    saveSettings();
  }
  else
  {
    motor.setConfig(settings.motorConfig);
    pedal.setConfig(settings.pedalConfig);
    Serial.println("Settings loaded from EEPROM.");
  }
}

void saveSettings()
{
  settings.motorConfig = motor.getConfig();
  settings.pedalConfig = pedal.getConfig();
  settings.checksum = calculateChecksum(settings);

  EEPROM.put(0, settings);
  Serial.println("Settings saved to EEPROM.");
}

uint8_t calculateChecksum(const Settings &s)
{
  uint8_t sum = 0;
  const uint8_t *data = reinterpret_cast<const uint8_t *>(&s);
  for (size_t i = 0; i < sizeof(Settings) - 1; i++)
  {
    sum ^= data[i];
  }
  return sum;
}

void handleEncoder()
{
  static uint8_t lastState = 0;
  uint8_t currentState = (digitalRead(encoderPinA) << 1) | digitalRead(encoderPinB);
  if ((lastState == 0b00 && currentState == 0b01) ||
      (lastState == 0b01 && currentState == 0b11) ||
      (lastState == 0b11 && currentState == 0b10) ||
      (lastState == 0b10 && currentState == 0b00))
  {
    // Increment
  }
  else if ((lastState == 0b00 && currentState == 0b10) ||
           (lastState == 0b10 && currentState == 0b11) ||
           (lastState == 0b11 && currentState == 0b01) ||
           (lastState == 0b01 && currentState == 0b00))
  {
    // Decrement
  }
  lastState = currentState;
}

void handleEncoderButton()
{
  // Handle encoder button press
}

void updateDisplay()
{
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);

  display.setCursor(0, 0);
  display.print("Pedal: ");
  display.println(pedal.getPedal());

  display.display();
}

void handleSerialInput()
{
  if (Serial.available())
  {
    char input = Serial.read();
    switch (input)
    {
    case 'c':
      Serial.println("Starting auto-calibration...");
      pedal.autoCalibrate(true);
      autoCalibrating = true;
      lastCalibrationState = IDLE; // Reset state tracking
      break;
    case 'x':
      Serial.println("Stopping auto-calibration or serial control.");
      autoCalibrating = false;
      serialControlEnabled = false;
      motor.stopMotor();
      break;
    case 'r':
      Serial.println("Loading defaults.");
      motor.loadDefaults();
      pedal.loadDefaults();
      break;
    case 's':
      Serial.println("Saving settings to EEPROM.");
      saveSettings();
      break;
    case 'l':
      Serial.println("Loading settings from EEPROM.");
      loadSettings();
      break;
    case 'h':
      printHelp();
      break;
    case '+':
      serialControlEnabled = true;
      motor.setTargetSpeed(motor.getSpeed() + 1);
      Serial.print("Increased Speed to: ");
      Serial.println(motor.getSpeed());
      break;
    case '-':
      serialControlEnabled = true;
      motor.setTargetSpeed(motor.getSpeed() - 1);
      Serial.print("Decreased Speed to: ");
      Serial.println(motor.getSpeed());
      break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      if (serialControlEnabled)
      {
        motor.setTargetSpeed((input - 0x30) * 10);
        Serial.print("Set Speed to: ");
        Serial.println(motor.getSpeed());
      }
      break;
    }
    Serial.println(input);
  }
}

void printHelp()
{
  Serial.println("Available commands:");
  Serial.println("c - Start auto-calibration");
  Serial.println("x - Stop auto-calibration or serial control");
  Serial.println("r - Load defaults");
  Serial.println("s - Save settings to EEPROM");
  Serial.println("l - Load settings from EEPROM");
  Serial.println("h - Show this help");
  Serial.println("+ - Increment PWM");
  Serial.println("- - Decrement PWM");
}

void configurePWMFrequency()
{
  // Configure Timer1 for higher PWM frequency (e.g., ~8 kHz)
  TCCR1B = (TCCR1B & 0b11111000) | 0x01; // Set prescaler to 1
}
